@@ -30,25 +30,30 @@
 #include <mach/machine.h>
 
 #include <vector>
-#include <ext/hash_set>
-#include <ext/hash_map>
+#include <unordered_set>
+#include <unordered_map>
 
 #include "ld.hpp"
+#include "Snapshot.h"
 
 extern void throwf (const char* format, ...) __attribute__ ((noreturn,format(printf, 1, 2)));
 extern void warning(const char* format, ...) __attribute__((format(printf, 1, 2)));
 
+class Snapshot;
+
 class LibraryOptions
 {
 public:
 	LibraryOptions() : fWeakImport(false), fReExport(false), fBundleLoader(false), 
-						fLazyLoad(false), fUpward(false), fForceLoad(false) {}
+						fLazyLoad(false), fUpward(false), fIndirectDylib(false), 
+						fForceLoad(false) {}
 	// for dynamic libraries
 	bool		fWeakImport;
 	bool		fReExport;
 	bool		fBundleLoader;
 	bool		fLazyLoad;
 	bool		fUpward;
+	bool		fIndirectDylib;
 	// for static libraries
 	bool		fForceLoad;
 };
@@ -82,11 +87,41 @@ public:
 	enum LocalSymbolHandling { kLocalSymbolsAll, kLocalSymbolsNone, kLocalSymbolsSelectiveInclude, kLocalSymbolsSelectiveExclude };
 	enum DebugInfoStripping { kDebugInfoNone, kDebugInfoMinimal, kDebugInfoFull };
 
-	struct FileInfo {
+	class FileInfo {
+    public:
 		const char*				path;
 		uint64_t				fileLen;
 		time_t					modTime;
 		LibraryOptions			options;
+		ld::File::Ordinal		ordinal;
+		bool					fromFileList;
+
+		// These are used by the threaded input file parsing engine.
+		mutable int				inputFileSlot;	// The input file "slot" assigned to this particular file
+		bool					readyToParse;
+
+        // The use pattern for FileInfo is to create one on the stack in a leaf function and return
+        // it to the calling frame by copy. Therefore the copy constructor steals the path string from
+        // the source, which dies with the stack frame.
+        FileInfo(FileInfo const &other) : path(other.path), fileLen(other.fileLen), modTime(other.modTime), options(other.options), ordinal(other.ordinal), fromFileList(other.fromFileList), inputFileSlot(-1) { ((FileInfo&)other).path = NULL; };
+
+        // Create an empty FileInfo. The path can be set implicitly by checkFileExists().
+        FileInfo() : path(NULL), fileLen(0), modTime(0), options(), fromFileList(false) {};
+        
+        // Create a FileInfo for a specific path, but does not stat the file.
+        FileInfo(const char *_path) : path(strdup(_path)), fileLen(0), modTime(0), options(), fromFileList(false) {};
+
+        ~FileInfo() { if (path) ::free((void*)path); }
+        
+        // Stat the file and update fileLen and modTime.
+        // If the object already has a path the p must be NULL.
+        // If the object does not have a path then p can be any candidate path, and if the file exists the object permanently remembers the path.
+        // Returns true if the file exists, false if not.
+        bool checkFileExists(const Options& options, const char *p=NULL);
+        
+        // Returns true if a previous call to checkFileExists() succeeded.
+        // Returns false if the file does not exist of checkFileExists() has never been called.
+        bool missing() const { return modTime==0; }
 	};
 
 	struct ExtraSection {
@@ -105,6 +140,11 @@ public:
 		uint8_t					alignment;
 	};
 
+	struct SectionOrderList {
+		const char*					segmentName;
+		std::vector<const char*>	sectionOrder;
+	};
+
 	struct OrderedSymbol {
 		const char*				symbolName;
 		const char*				objectFileName;
@@ -137,6 +177,25 @@ public:
 		const char*			alias;
 	};
 
+	struct SectionRename {
+		const char*			fromSegment;
+		const char*			fromSection;
+		const char*			toSegment;
+		const char*			toSection;
+	};
+
+	struct SegmentRename {
+		const char*			fromSegment;
+		const char*			toSegment;
+	};
+
+	enum { depLinkerVersion=0x00, depObjectFile=0x10, depDirectDylib=0x10, depIndirectDylib=0x10, 
+		  depUpwardDirectDylib=0x10, depUpwardIndirectDylib=0x10, depArchive=0x10,
+		  depFileList=0x10, depSection=0x10, depBundleLoader=0x10, depMisc=0x10, depNotFound=0x11,
+		  depOutputFile = 0x40 };
+	
+	void						dumpDependency(uint8_t, const char* path) const;
+	
 	typedef const char* const*	UndefinesIterator;
 
 //	const ObjectFile::ReaderOptions&	readerOptions();
@@ -156,7 +215,8 @@ public:
 	bool						bindAtLoad() const { return fBindAtLoad; }
 	NameSpace					nameSpace() const { return fNameSpace; }
 	const char*					installPath() const;			// only for kDynamicLibrary
-	uint32_t					currentVersion() const { return fDylibCurrentVersion; }		// only for kDynamicLibrary
+	uint64_t					currentVersion() const { return fDylibCurrentVersion; }		// only for kDynamicLibrary
+	uint32_t					currentVersion32() const;		// only for kDynamicLibrary
 	uint32_t					compatibilityVersion() const { return fDylibCompatVersion; }	// only for kDynamicLibrary
 	const char*					entryName() const { return fEntryName; }		// only for kDynamicExecutable or kStaticExecutable
 	const char*					executablePath();
@@ -197,6 +257,7 @@ public:
 	bool						hasNonExecutableHeap() const { return fNonExecutableHeap; }
 	UndefinesIterator			initialUndefinesBegin() const { return &fInitialUndefines[0]; }
 	UndefinesIterator			initialUndefinesEnd() const { return &fInitialUndefines[fInitialUndefines.size()]; }
+	const std::vector<const char*>&	initialUndefines() const { return fInitialUndefines; }
 	bool						printWhyLive(const char* name) const;
 	uint32_t					minimumHeaderPad() const { return fMinimumHeaderPad; }
 	bool						maxMminimumHeaderPad() const { return fMaxMinimumHeaderPad; }
@@ -243,12 +304,16 @@ public:
 	bool						keepLocalSymbol(const char* symbolName) const;
 	bool						allowTextRelocs() const { return fAllowTextRelocs; }
 	bool						warnAboutTextRelocs() const { return fWarnTextRelocs; }
+	bool						kextsUseStubs() const { return fKextsUseStubs; }
 	bool						usingLazyDylibLinking() const { return fUsingLazyDylibLinking; }
 	bool						verbose() const { return fVerbose; }
 	bool						makeEncryptable() const { return fEncryptable; }
 	bool						needsUnwindInfoSection() const { return fAddCompactUnwindEncoding; }
 	const std::vector<const char*>&	llvmOptions() const{ return fLLVMOptions; }
+	const std::vector<const char*>&	segmentOrder() const{ return fSegmentOrder; }
+	const std::vector<const char*>* sectionOrder(const char* segName) const;
 	const std::vector<const char*>&	dyldEnvironExtras() const{ return fDyldEnvironExtras; }
+	const std::vector<const char*>&	astFilePaths() const{ return fASTFilePaths; }
 	bool						makeCompressedDyldInfo() const { return fMakeCompressedDyldInfo; }
 	bool						hasExportedSymbolOrder();
 	bool						exportedSymbolOrder(const char* sym, unsigned int* order) const;
@@ -281,27 +346,58 @@ public:
 	bool						objcGc() const { return fObjCGc; }
 	bool						objcGcOnly() const { return fObjCGcOnly; }
 	bool						canUseThreadLocalVariables() const { return fTLVSupport; }
-	bool						demangleSymbols() const { return fDemangle; }
 	bool						addVersionLoadCommand() const { return fVersionLoadCommand; }
 	bool						addFunctionStarts() const { return fFunctionStartsLoadCommand; }
+	bool						addDataInCodeInfo() const { return fDataInCodeInfoLoadCommand; }
 	bool						canReExportSymbols() const { return fCanReExportSymbols; }
 	const char*					tempLtoObjectPath() const { return fTempLtoObjectPath; }
+	const char*					overridePathlibLTO() const { return fOverridePathlibLTO; }
+	const char*					mcpuLTO() const { return fLtoCpu; }
 	bool						objcCategoryMerging() const { return fObjcCategoryMerging; }
 	bool						pageAlignDataAtoms() const { return fPageAlignDataAtoms; }
+	bool						keepDwarfUnwind() const { return fKeepDwarfUnwind; }
+	bool						verboseOptimizationHints() const { return fVerboseOptimizationHints; }
+	bool						ignoreOptimizationHints() const { return fIgnoreOptimizationHints; }
+	bool						generateDtraceDOF() const { return fGenerateDtraceDOF; }
+	bool						allowBranchIslands() const { return fAllowBranchIslands; }
+	bool						traceSymbolLayout() const { return fTraceSymbolLayout; }
+	bool						markAppExtensionSafe() const { return fMarkAppExtensionSafe; }
+	bool						checkDylibsAreAppExtensionSafe() const { return fCheckAppExtensionSafe; }
+	bool						forceLoadSwiftLibs() const { return fForceLoadSwiftLibs; }
 	bool						hasWeakBitTweaks() const;
 	bool						forceWeak(const char* symbolName) const;
 	bool						forceNotWeak(const char* symbolName) const;
 	bool						forceWeakNonWildCard(const char* symbolName) const;
 	bool						forceNotWeakNonWildcard(const char* symbolName) const;
+	bool						forceCoalesce(const char* symbolName) const;
+    Snapshot&                   snapshot() const { return fLinkSnapshot; }
+	bool						errorBecauseOfWarnings() const;
+	bool						needsThreadLoadCommand() const { return fNeedsThreadLoadCommand; }
+	bool						needsEntryPointLoadCommand() const { return fEntryPointLoadCommand; }
+	bool						needsSourceVersionLoadCommand() const { return fSourceVersionLoadCommand; }
+	bool						needsDependentDRInfo() const { return fDependentDRInfo; }
+	bool						canUseAbsoluteSymbols() const { return fAbsoluteSymbols; }
+	bool						allowSimulatorToLinkWithMacOSX() const { return fAllowSimulatorToLinkWithMacOSX; }
+	uint64_t					sourceVersion() const { return fSourceVersion; }
+	uint32_t					sdkVersion() const { return fSDKVersion; }
+	const char*					demangleSymbol(const char* sym) const;
+    bool						pipelineEnabled() const { return fPipelineFifo != NULL; }
+    const char*					pipelineFifo() const { return fPipelineFifo; }
+	bool						dumpDependencyInfo() const { return (fDependencyInfoPath != NULL); }
+	const char*					dependencyInfoPath() const { return fDependencyInfoPath; }
+	bool						targetIOSSimulator() const { return fTargetIOSSimulator; }
+	ld::relocatable::File::LinkerOptionsList&	
+								linkerOptions() const { return fLinkerOptions; }
+	FileInfo					findFramework(const char* frameworkName) const;
+	FileInfo					findLibrary(const char* rootName, bool dylibsOnly=false) const;
+	const std::vector<SectionRename>& sectionRenames() const { return fSectionRenames; }
+	const std::vector<SegmentRename>& segmentRenames() const { return fSegmentRenames; }
+	bool						moveRoSymbol(const char* symName, const char* filePath, const char*& seg, bool& wildCardMatch) const;
+	bool						moveRwSymbol(const char* symName, const char* filePath, const char*& seg, bool& wildCardMatch) const;
 
 private:
-	class CStringEquals
-	{
-	public:
-		bool operator()(const char* left, const char* right) const { return (strcmp(left, right) == 0); }
-	};
-	typedef __gnu_cxx::hash_map<const char*, unsigned int, __gnu_cxx::hash<const char*>, CStringEquals> NameToOrder;
-	typedef __gnu_cxx::hash_set<const char*, __gnu_cxx::hash<const char*>, CStringEquals>  NameSet;
+	typedef std::unordered_map<const char*, unsigned int, ld::CStringHash, ld::CStringEquals> NameToOrder;
+	typedef std::unordered_set<const char*, ld::CStringHash, ld::CStringEquals>  NameSet;
 	enum ExportMode { kExportDefault, kExportSome, kDontExportSome };
 	enum LibrarySearchMode { kSearchDylibAndArchiveInEachDir, kSearchAllDirsForDylibsThenAllDirsForArchives };
 	enum InterposeMode { kInterposeNone, kInterposeAllExternal, kInterposeSome };
@@ -309,7 +405,8 @@ private:
 	class SetWithWildcards {
 	public:
 		void					insert(const char*);
-		bool					contains(const char*) const;
+		bool					contains(const char*, bool* wildCardMatch=NULL) const;
+		bool					containsWithPrefix(const char* symbol, const char* file, bool& wildCardMatch) const;
 		bool					containsNonWildcard(const char*) const;
 		bool					empty() const			{ return fRegular.empty() && fWildCard.empty(); }
 		bool					hasWildCards() const	{ return !fWildCard.empty(); }
@@ -325,22 +422,25 @@ private:
 		std::vector<const char*>		fWildCard;
 	};
 
+	struct SymbolsMove {
+		const char*			toSegment;
+		SetWithWildcards	symbols;
+	};
 
 	void						parse(int argc, const char* argv[]);
 	void						checkIllegalOptionCombinations();
 	void						buildSearchPaths(int argc, const char* argv[]);
 	void						parseArch(const char* architecture);
-	FileInfo					findLibrary(const char* rootName, bool dylibsOnly=false);
-	FileInfo					findFramework(const char* frameworkName);
-	FileInfo					findFramework(const char* rootName, const char* suffix);
+	FileInfo					findFramework(const char* rootName, const char* suffix) const;
 	bool						checkForFile(const char* format, const char* dir, const char* rootName,
 											 FileInfo& result) const;
-	uint32_t					parseVersionNumber(const char*);
+	uint64_t					parseVersionNumber64(const char*);
+	uint32_t					parseVersionNumber32(const char*);
 	void						parseSectionOrderFile(const char* segment, const char* section, const char* path);
 	void						parseOrderFile(const char* path, bool cstring);
 	void						addSection(const char* segment, const char* section, const char* path);
 	void						addSubLibrary(const char* name);
-	void						loadFileList(const char* fileOfPaths);
+	void						loadFileList(const char* fileOfPaths, ld::File::Ordinal baseOrdinal);
 	uint64_t					parseAddress(const char* addr);
 	void						loadExportFile(const char* fileOfExports, const char* option, SetWithWildcards& set);
 	void						parseAliasFile(const char* fileOfAliases);
@@ -361,7 +461,9 @@ private:
 	void						warnObsolete(const char* arg);
 	uint32_t					parseProtection(const char* prot);
 	void						loadSymbolOrderFile(const char* fileOfExports, NameToOrder& orderMapping);
-
+	void						addSectionRename(const char* srcSegment, const char* srcSection, const char* dstSegment, const char* dstSection);
+	void						addSegmentRename(const char* srcSegment, const char* dstSegment);
+	void						addSymbolMove(const char* dstSegment, const char* symbolList, std::vector<SymbolsMove>& list, const char* optionName);
 
 
 //	ObjectFile::ReaderOptions			fReaderOptions;
@@ -384,7 +486,7 @@ private:
 	bool								fDeadStrip;
 	NameSpace							fNameSpace;
 	uint32_t							fDylibCompatVersion;
-	uint32_t							fDylibCurrentVersion;
+	uint64_t							fDylibCurrentVersion;
 	const char*							fDylibInstallName;
 	const char*							fFinalName;
 	const char*							fEntryName;
@@ -398,6 +500,7 @@ private:
 	SetWithWildcards					fForceWeakSymbols;
 	SetWithWildcards					fForceNotWeakSymbols;
 	SetWithWildcards					fReExportSymbols;
+	SetWithWildcards					fForceCoalesceSymbols;
 	NameSet								fRemovedExports;
 	NameToOrder							fExportSymbolsOrder;
 	ExportMode							fExportMode;
@@ -420,9 +523,13 @@ private:
 	const char*							fMapPath;
 	const char*							fDyldInstallPath;
 	const char*							fTempLtoObjectPath;
+	const char*							fOverridePathlibLTO;
+	const char*							fLtoCpu;
 	uint64_t							fZeroPageSize;
 	uint64_t							fStackSize;
 	uint64_t							fStackAddr;
+	uint64_t							fSourceVersion;
+	uint32_t							fSDKVersion;
 	bool								fExecutableStack;
 	bool								fNonExecutableHeap;
 	bool								fDisableNonExecutableHeap;
@@ -451,6 +558,7 @@ private:
 	bool								fDeadStripDylibs;
 	bool								fAllowTextRelocs;
 	bool								fWarnTextRelocs;
+	bool								fKextsUseStubs;
 	bool								fUsingLazyDylibLinking;
 	bool								fEncryptable;
 	bool								fOrderData;
@@ -499,9 +607,37 @@ private:
 	bool								fFunctionStartsLoadCommand;
 	bool								fFunctionStartsForcedOn;
 	bool								fFunctionStartsForcedOff;
+	bool								fDataInCodeInfoLoadCommand;
+	bool								fDataInCodeInfoLoadCommandForcedOn;
+	bool								fDataInCodeInfoLoadCommandForcedOff;
 	bool								fCanReExportSymbols;
 	bool								fObjcCategoryMerging;
 	bool								fPageAlignDataAtoms;
+	bool								fNeedsThreadLoadCommand;
+	bool								fEntryPointLoadCommand;
+	bool								fEntryPointLoadCommandForceOn;
+	bool								fEntryPointLoadCommandForceOff;
+	bool								fSourceVersionLoadCommand;
+	bool								fSourceVersionLoadCommandForceOn;
+	bool								fSourceVersionLoadCommandForceOff;	
+	bool								fDependentDRInfo;
+	bool								fDependentDRInfoForcedOn;
+	bool								fDependentDRInfoForcedOff;
+	bool								fTargetIOSSimulator;
+	bool								fExportDynamic;
+	bool								fAbsoluteSymbols;
+	bool								fAllowSimulatorToLinkWithMacOSX;
+	bool								fKeepDwarfUnwind;
+	bool								fKeepDwarfUnwindForcedOn;
+	bool								fKeepDwarfUnwindForcedOff;
+	bool								fVerboseOptimizationHints;
+	bool								fIgnoreOptimizationHints;
+	bool								fGenerateDtraceDOF;
+	bool								fAllowBranchIslands;
+	bool								fTraceSymbolLayout;
+	bool								fMarkAppExtensionSafe;
+	bool								fCheckAppExtensionSafe;
+	bool								fForceLoadSwiftLibs;
 	DebugInfoStripping					fDebugInfoStripping;
 	const char*							fTraceOutputFile;
 	ld::MacVersionMin					fMacVersionMin;
@@ -522,7 +658,21 @@ private:
 	std::vector<const char*>			fFrameworkSearchPaths;
 	std::vector<const char*>			fSDKPaths;
 	std::vector<const char*>			fDyldEnvironExtras;
+	std::vector<const char*>			fSegmentOrder;
+	std::vector<const char*>			fASTFilePaths;
+	std::vector<SectionOrderList>		fSectionOrder;
+	std::vector< std::vector<const char*> > fLinkerOptions;
+	std::vector<SectionRename>			fSectionRenames;
+	std::vector<SegmentRename>			fSegmentRenames;
+	std::vector<SymbolsMove>			fSymbolsMovesData;
+	std::vector<SymbolsMove>			fSymbolsMovesCode;
+	std::vector<SymbolsMove>			fSymbolsMovesZeroFill;
 	bool								fSaveTempFiles;
+    mutable Snapshot					fLinkSnapshot;
+    bool								fSnapshotRequested;
+    const char*							fPipelineFifo;
+	const char*							fDependencyInfoPath;
+	mutable int							fDependencyFileDescriptor;
 };
 
 
